#ifndef GML_H
#define GML_H

#include <hashtable.h>
#include <list.h>
#include <math.h>
#include <unixlib.h>
#include <core.h>
#include <string.h>
#include <char.h>

open List;
open Hashtable;
open Core;
open Math;
open String;
open Char;


#define env <*(string,value)>list 
#define op_t string,*(value,<value>list) f(<value>list)
#define PI 3.1415927f

union <a>option {
    a SOME;
    void NONE;
}

union base_value {
    bool Bool;
    double Real;
    int Int;
    string String;
    void Unit;
}

union value {
    base_value BaseValue;
    <value>list Array;
    object Object;
    light Light;
    point Point;
    color Color;
    *(env,<token_group>list) Closure;
}

union token {
    string Operator;
    string Identifier;
    string Binder;
    value Value;
    void LBRACE;
    void RBRACE;
    void LBRACKET;
    void RBRACKET;
}

union token_group {
    token Token;
    <token_group>list Function;
    <token_group>list Array;
}


struct machine {
    env e;
    <value>list a;
    <token_group>list c;
}

extern <op_t>table setup_ops();


/* SCENE STUFF */

struct point {
    double x,y,z;
}

/* invariant: dir always normalized */
struct ray {
    point orig;
    point dir;
}

struct color {
    double r,g,b;
}


union surface {
    *(env,<token_group>list) Fun;
    /* diffuse, specular, phong;  face, u, v, closure stuff */
    *(color,double,double,double) Const;
}


struct homogenous {
  double x, y, z, w;
}

struct matrix4 {
  homogenous r0;
  homogenous r1;
  homogenous r2;
  homogenous r3;
}

union transf {
    point Translate;
    point Scale;
    double RotX;
    double RotY;
    double RotZ;
}

union transform {
    <transf>list Transf;
    *(matrix4,matrix4) Matrix;
}

union solid {
    *(transform,surface) Sphere;  
    *(transform,surface) Plane;   
    *(transform,surface) Cone;
    *(transform,surface) Cylinder;
    /* Cube is just the intersection of 6 planes */
}

union object {
    solid Solid;
    *(object,object) Union;
    *(object,object) Intersect;
    *(object,object) CIntersect;
    *(object,object) Difference;
}

union light {
    *(point,color) Directional;     /* direction */
    *(point,color) Point;           /* position */
    *(point,point,color,double,double) Spot;
    /* position, pointing at, color, cutoff angle, attenuation exponent */
}

struct scene {
    color amb;
    <light>list lights;
    object obj;
    int depth;
    double fov;
    int width;
    int height;
    string file;
}

extern machine evaluate(machine m, <op_t>table ops);


extern exception OpErrorPrime(string);
extern exception EvaluError(string);
extern exception ParseError(string);
extern exception OpError(string);
extern exception TypeError(string);
extern exception BadTransf;
extern exception Unimplemented(string);
extern exception BadTransform;


#endif
