#ifndef SCENE_H
#define SCENE_H

#include "gml.h"


struct point {
  double x,y,z;
}

/* invariant: dir always normalized */
struct ray {
  point orig;
  point dir;
}

struct color {
  char r,g,b;
}


struct surface {
  *(color,double,double,double) f(int,double,double,env,<token_group>list);
  /* diffuse, specular, phong;  face, u, v */
}

union transform {
  point Translate;
  point Scale;
  double RotX;
  double RotY;
  double RotZ;
}

union solid {
  *(<transform>list,surface,env,<token_group>list) Sphere;  
  *(<transform>list,surface) Plane;   
  *(<transform>list,surface) Cube;    
}

union object {
  solid Solid;
  *(solid,solid) Union;
  *(solid,solid) Intersect;
}

union light {
  *(point,color) Directional;     /* direction */
  *(point,color) Point;           /* position */
}

struct scene {
  color amb;
  <light>list lights;
  <object>list objs;
  int depth;
  double fov;
  int width;
  int height;
  string file;
}

#endif
